"""Unit tests for custom exceptions."""

import unittest

from dtmapi.exceptions import (
    DTMApiError,
    DTMApiRequestError,
    DTMApiResponseError,
    DTMApiTimeoutError,
    DTMAuthenticationError,
)


class TestExceptions(unittest.TestCase):
    """Test custom exception classes."""

    def test_base_exception(self):
        """Test DTMApiError base exception."""
        with self.assertRaises(DTMApiError):
            raise DTMApiError("Base error")

    def test_authentication_error(self):
        """Test DTMAuthenticationError."""
        with self.assertRaises(DTMAuthenticationError) as context:
            raise DTMAuthenticationError("Auth failed")
        self.assertEqual(str(context.exception), "Auth failed")
        self.assertIsInstance(context.exception, DTMApiError)

    def test_response_error(self):
        """Test DTMApiResponseError."""
        error_messages = ["Error 1", "Error 2"]
        with self.assertRaises(DTMApiResponseError) as context:
            raise DTMApiResponseError("Response error", error_messages)
        self.assertEqual(str(context.exception), "Response error")
        self.assertEqual(context.exception.error_messages, error_messages)
        self.assertIsInstance(context.exception, DTMApiError)

    def test_response_error_without_messages(self):
        """Test DTMApiResponseError without error messages."""
        with self.assertRaises(DTMApiResponseError) as context:
            raise DTMApiResponseError("Response error")
        self.assertEqual(context.exception.error_messages, [])

    def test_request_error(self):
        """Test DTMApiRequestError."""
        with self.assertRaises(DTMApiRequestError) as context:
            raise DTMApiRequestError("Request failed")
        self.assertEqual(str(context.exception), "Request failed")
        self.assertIsInstance(context.exception, DTMApiError)

    def test_timeout_error(self):
        """Test DTMApiTimeoutError."""
        with self.assertRaises(DTMApiTimeoutError) as context:
            raise DTMApiTimeoutError("Timeout")
        self.assertEqual(str(context.exception), "Timeout")
        self.assertIsInstance(context.exception, DTMApiError)

    def test_exception_inheritance(self):
        """Test that all custom exceptions inherit from DTMApiError."""
        exceptions = [
            DTMAuthenticationError("test"),
            DTMApiResponseError("test"),
            DTMApiRequestError("test"),
            DTMApiTimeoutError("test"),
        ]
        for exc in exceptions:
            self.assertIsInstance(exc, DTMApiError)


if __name__ == "__main__":
    unittest.main()



"""Custom exceptions for the DTM API client."""


class DTMApiError(Exception):
    """Base exception for DTM API errors."""

    pass


class DTMAuthenticationError(DTMApiError):
    """Raised when authentication with the API fails."""

    pass


class DTMApiResponseError(DTMApiError):
    """Raised when the API returns an error response."""

    def __init__(self, message: str, error_messages: list = None):
        """
        Initialize the exception.

        :param message: The error message.
        :type message: str
        :param error_messages: List of error messages from the API response.
        :type error_messages: list
        """
        super().__init__(message)
        self.error_messages = error_messages or []


class DTMApiRequestError(DTMApiError):
    """Raised when a request to the API fails."""

    pass


class DTMApiTimeoutError(DTMApiError):
    """Raised when a request to the API times out."""

    pass


class DTMApiVersionError(DTMApiError):
    """Raised when an invalid API version is specified."""

    pass



import logging
import os
import time
from typing import Any, Dict, Optional, Union

import pandas as pd
import requests

from dtmapi.config import (
    COUNTRY_LIST_API,
    COUNTRY_LIST_API_V2,
    IDP_ADMIN_0_API,
    IDP_ADMIN_0_API_V2,
    IDP_ADMIN_1_API,
    IDP_ADMIN_1_API_V2,
    IDP_ADMIN_2_API,
    IDP_ADMIN_2_API_V2,
    OPERATION_LIST_API,
    OPERATION_LIST_API_V2,
)
from dtmapi.exceptions import (
    DTMApiRequestError,
    DTMApiResponseError,
    DTMApiTimeoutError,
    DTMAuthenticationError,
    DTMApiVersionError,
)
from dtmapi.validators import (
    ValidationError,
    validate_date_format,
    validate_date_range,
    validate_required_params,
    validate_round_number_range,
)

# Configure logger
logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


class DTMApi:
    """
    Python interface to DTM API endpoints. Supports both v2 (legacy) and v3 (current).
    Requires Ocp-Apim-Subscription-Key for authentication.

    **API Version Differences:**

    +-------------------------------+----------+----------+
    | Feature                       | v2       | v3       |
    +===============================+==========+==========+
    | IDP Admin 0, 1, 2 Data        | Yes      | Yes      |
    +-------------------------------+----------+----------+
    | Country List                  | Yes      | Yes      |
    +-------------------------------+----------+----------+
    | Operation List                | Yes      | Yes      |
    +-------------------------------+----------+----------+
    | Gender/Sex Disaggregation     | No       | **Yes**  |
    +-------------------------------+----------+----------+
    | Origin of Displacement        | No       | **Yes**  |
    +-------------------------------+----------+----------+
    | Displacement Reason           | No       | **Yes**  |
    +-------------------------------+----------+----------+
    | Status                        | Legacy   | Current  |
    +-------------------------------+----------+----------+

    **Usage Examples:**

    v3 API (default - recommended for new projects)::

        api = DTMApi(subscription_key="YOUR-KEY")
        data = api.get_idp_admin0_data(CountryName="Sudan")
        # Returns IDP data WITH gender, origin, and reason fields

    v2 API (legacy - for historical data compatibility)::

        api = DTMApi(subscription_key="YOUR-KEY", api_version="v2")
        data = api.get_idp_admin0_data(CountryName="Sudan")
        # Returns IDP data WITHOUT demographic disaggregation
    """

    DEFAULT_TIMEOUT = 30  # seconds
    DEFAULT_MAX_RETRIES = 3
    DEFAULT_RETRY_DELAY = 1  # seconds

    def __init__(
        self,
        subscription_key: Optional[str] = None,
        api_version: str = "v3",
        timeout: Optional[int] = None,
        max_retries: Optional[int] = None,
        retry_delay: Optional[float] = None,
    ):
        """
        Initialize the DTM API client.

        :param subscription_key: DTM API subscription key. Can also be set via DTMAPI_SUBSCRIPTION_KEY env var.
        :type subscription_key: Optional[str]
        :param api_version: API version to use: "v2" (legacy) or "v3" (current, default).
        :type api_version: str
        :param timeout: Request timeout in seconds (default: 30).
        :type timeout: Optional[int]
        :param max_retries: Maximum number of retry attempts for failed requests (default: 3).
        :type max_retries: Optional[int]
        :param retry_delay: Base delay in seconds between retries (default: 1). Uses exponential backoff.
        :type retry_delay: Optional[float]
        :raises DTMAuthenticationError: If no subscription key is provided.
        :raises DTMApiVersionError: If api_version is not "v2" or "v3".
        """
        logger.debug("Initializing DTMApi client")

        # Validate API version
        if api_version not in ["v2", "v3"]:
            raise DTMApiVersionError(
                f"api_version must be 'v2' or 'v3', got: '{api_version}'"
            )
        self.api_version = api_version

        # Validate and set subscription key
        self.subscription_key = subscription_key or os.getenv("DTMAPI_SUBSCRIPTION_KEY")
        if not self.subscription_key:
            logger.error("No subscription key provided")
            raise DTMAuthenticationError(
                "A DTM API Subscription Key is required. "
                "Provide it as an argument or set the DTMAPI_SUBSCRIPTION_KEY environment variable."
            )

        # Set timeout and retry parameters
        self.timeout = timeout or self.DEFAULT_TIMEOUT
        self.max_retries = max_retries if max_retries is not None else self.DEFAULT_MAX_RETRIES
        self.retry_delay = retry_delay if retry_delay is not None else self.DEFAULT_RETRY_DELAY

        logger.info(
            f"DTMApi client initialized with api_version={self.api_version}, "
            f"timeout={self.timeout}s, max_retries={self.max_retries}, "
            f"retry_delay={self.retry_delay}s"
        )

    def _headers(self) -> Dict[str, str]:
        return {
            "User-Agent": "Mozilla/5.0 (compatible; DTMClient/2.0)",
            "Ocp-Apim-Subscription-Key": self.subscription_key,
        }

    def _get_endpoint(self, endpoint_type: str) -> str:
        """
        Get the appropriate API endpoint URL based on the API version.

        :param endpoint_type: Type of endpoint ("admin0", "admin1", "admin2", "countries", "operations").
        :type endpoint_type: str
        :return: The full API endpoint URL.
        :rtype: str
        """
        endpoint_map = {
            "v3": {
                "admin0": IDP_ADMIN_0_API,
                "admin1": IDP_ADMIN_1_API,
                "admin2": IDP_ADMIN_2_API,
                "countries": COUNTRY_LIST_API,
                "operations": OPERATION_LIST_API,
            },
            "v2": {
                "admin0": IDP_ADMIN_0_API_V2,
                "admin1": IDP_ADMIN_1_API_V2,
                "admin2": IDP_ADMIN_2_API_V2,
                "countries": COUNTRY_LIST_API_V2,
                "operations": OPERATION_LIST_API_V2,
            },
        }
        return endpoint_map[self.api_version][endpoint_type]

    def _is_retryable_error(self, exception: Exception) -> bool:
        """
        Determine if an error should trigger a retry.

        :param exception: The exception to check.
        :type exception: Exception
        :return: True if the error is retryable, False otherwise.
        :rtype: bool
        """
        # Retry on timeout
        if isinstance(exception, requests.Timeout):
            return True

        # Retry on connection errors
        if isinstance(exception, requests.ConnectionError):
            return True

        # Retry on specific HTTP status codes (429, 500, 502, 503, 504)
        if isinstance(exception, requests.HTTPError):
            if exception.response is not None:
                status_code = exception.response.status_code
                # Rate limiting or server errors
                if status_code in [429, 500, 502, 503, 504]:
                    return True

        return False

    def _fetch_data(
        self,
        api_url: str,
        params: Optional[Dict[str, Any]] = None,
        to_pandas: bool = True,
    ) -> Union[pd.DataFrame, Dict[str, Any]]:
        """
        Helper function to fetch data from the specified API URL with given parameters.
        Implements retry logic with exponential backoff for transient errors.

        :param api_url: The API endpoint URL.
        :type api_url: str
        :param params: The query parameters for the API request.
        :type params: Dict[str, Any]
        :param to_pandas: If True, the data will be returned as a pandas DataFrame. Otherwise, it will be returned as a JSON object.
        :type to_pandas: bool
        :return: The data matching the specified criteria, either as a DataFrame or a JSON object.
        :rtype: Union[pd.DataFrame, Dict[str, Any]]
        :raises DTMApiTimeoutError: If the request times out.
        :raises DTMApiAuthenticationError: If authentication fails.
        :raises DTMApiResponseError: If the API returns an error response.
        :raises DTMApiRequestError: If the request fails for other reasons.
        """
        logger.debug(f"Fetching data from {api_url} with params={params}")
        last_exception = None

        for attempt in range(self.max_retries + 1):
            try:
                if attempt > 0:
                    # Exponential backoff: delay * (2 ^ (attempt - 1))
                    delay = self.retry_delay * (2 ** (attempt - 1))
                    logger.info(f"Retrying request (attempt {attempt + 1}/{self.max_retries + 1}) after {delay}s delay")
                    time.sleep(delay)

                response = requests.get(
                    api_url, params=params, headers=self._headers(), timeout=self.timeout
                )
                logger.debug(f"Received response: status={response.status_code}")
                response.raise_for_status()
                data = response.json()

                # For endpoints that wrap result in 'isSuccess'
                if not data.get("isSuccess"):
                    error_messages = data.get("errorMessages", [])
                    error_message = (
                        error_messages[0] if error_messages else "Unknown API error"
                    )
                    logger.error(f"API returned error: {error_message}")
                    raise DTMApiResponseError(error_message, error_messages)

                result = data["result"]
                result_count = len(result) if isinstance(result, list) else 1
                logger.info(f"Successfully fetched {result_count} records from {api_url}")
                return pd.DataFrame(result) if to_pandas else result

            except requests.Timeout as e:
                last_exception = e
                logger.warning(f"Request timed out after {self.timeout} seconds: {api_url}")
                if not self._is_retryable_error(e) or attempt >= self.max_retries:
                    logger.error(f"Request timed out after {self.timeout} seconds (no more retries)")
                    raise DTMApiTimeoutError(
                        f"Request timed out after {self.timeout} seconds"
                    ) from e
            except requests.HTTPError as e:
                last_exception = e
                if e.response.status_code == 401 or e.response.status_code == 403:
                    logger.error(f"Authentication failed: {e.response.status_code}")
                    raise DTMAuthenticationError(
                        f"Authentication failed: {e.response.status_code} {e.response.reason}"
                    ) from e
                logger.warning(f"HTTP error: {e.response.status_code} {e.response.reason}")
                if not self._is_retryable_error(e) or attempt >= self.max_retries:
                    logger.error(f"HTTP error occurred (no more retries): {e.response.status_code}")
                    raise DTMApiRequestError(
                        f"HTTP error occurred: {e.response.status_code} {e.response.reason}"
                    ) from e
            except requests.RequestException as e:
                last_exception = e
                logger.warning(f"Request failed: {e}")
                if not self._is_retryable_error(e) or attempt >= self.max_retries:
                    logger.error(f"Request failed (no more retries): {e}")
                    raise DTMApiRequestError(f"API request failed: {e}") from e

        # This should not be reached, but just in case
        if last_exception:
            raise DTMApiRequestError(f"API request failed after {self.max_retries} retries") from last_exception

    # ----------- Public API Methods -----------

    def get_all_countries(
        self, to_pandas: bool = True
    ) -> Union[pd.DataFrame, Dict[str, Any]]:
        """
        Retrieve all countries for which DTM data is publicly available through the API.

        :return: All countries for which DTM data is publicly available through the API.
        :rtype: Union[pd.DataFrame, Dict[str, Any]]
        """
        return self._fetch_data(self._get_endpoint("countries"), to_pandas=to_pandas)

    def get_all_operations(
        self, to_pandas: bool = True
    ) -> Union[pd.DataFrame, Dict[str, Any]]:
        """
        Retrieve all operations for which DTM data is publicly available through the API.

        :return: All operations for which DTM data is publicly available through the API.
        :rtype: Union[pd.DataFrame, Dict[str, Any]]
        """
        return self._fetch_data(self._get_endpoint("operations"), to_pandas=to_pandas)

    def get_idp_admin0_data(
        self,
        Operation: Optional[str] = None,
        CountryName: Optional[str] = None,
        Admin0Pcode: Optional[str] = None,
        FromReportingDate: Optional[str] = None,
        ToReportingDate: Optional[str] = None,
        FromRoundNumber: Optional[int] = None,
        ToRoundNumber: Optional[int] = None,
        to_pandas: bool = True,
    ) -> Union[pd.DataFrame, Dict[str, Any]]:
        """
        Retrieve IDP data at Admin 0 level based on specified parameters.

        At least one of the following parameters must be provided:
        Operation, CountryName, or Admin0Pcode.

        :param Operation: Name of the DTM operation for which the data was collected.
        :type Operation: Optional[str]
        :param CountryName: Name of the country where the data was collected.
        :type CountryName: Optional[str]
        :param Admin0Pcode: Country code (ISO 3166-1 alpha-3).
        :type Admin0Pcode: Optional[str]
        :param FromReportingDate: Start date for the reporting period (format: 'YYYY-MM-DD').
        :type FromReportingDate: Optional[str]
        :param ToReportingDate: End date for the reporting period (format: 'YYYY-MM-DD').
        :type ToReportingDate: Optional[str]
        :param FromRoundNumber: Starting round number for the data collection range.
        :type FromRoundNumber: Optional[int]
        :param ToRoundNumber: Ending round number for the data collection range.
        :type ToRoundNumber: Optional[int]
        :param to_pandas: If True, the data will be returned as a pandas DataFrame. Otherwise, it will be returned as a JSON object.
        :type to_pandas: bool

        :return: The IDP Admin0 data matching the specified criteria, either as a DataFrame or a JSON object.
        :rtype: Union[pd.DataFrame, Dict[str, Any]]
        :raises ValidationError: If parameter validation fails.
        """
        # Validate required parameters
        validate_required_params(
            {"Operation": Operation, "CountryName": CountryName, "Admin0Pcode": Admin0Pcode},
            ["Operation", "CountryName", "Admin0Pcode"]
        )

        # Validate date formats
        if FromReportingDate:
            validate_date_format(FromReportingDate, "FromReportingDate")
        if ToReportingDate:
            validate_date_format(ToReportingDate, "ToReportingDate")
        validate_date_range(FromReportingDate, ToReportingDate, "FromReportingDate", "ToReportingDate")

        # Validate round numbers
        validate_round_number_range(FromRoundNumber, ToRoundNumber)

        params = {
            "Operation": Operation,
            "CountryName": CountryName,
            "Admin0Pcode": Admin0Pcode,
            "FromReportingDate": FromReportingDate,
            "ToReportingDate": ToReportingDate,
            "FromRoundNumber": FromRoundNumber,
            "ToRoundNumber": ToRoundNumber,
        }
        # Remove None values
        params = {k: v for k, v in params.items() if v is not None}
        return self._fetch_data(self._get_endpoint("admin0"), params, to_pandas)

    def get_idp_admin1_data(
        self,
        Operation: Optional[str] = None,
        CountryName: Optional[str] = None,
        Admin0Pcode: Optional[str] = None,
        Admin1Name: Optional[str] = None,
        Admin1Pcode: Optional[str] = None,
        FromReportingDate: Optional[str] = None,
        ToReportingDate: Optional[str] = None,
        FromRoundNumber: Optional[int] = None,
        ToRoundNumber: Optional[int] = None,
        to_pandas: bool = True,
    ) -> Union[pd.DataFrame, Dict[str, Any]]:
        """
        Retrieve IDP data at Admin 1 level based on specified parameters.

        At least one of the following parameters must be provided:
        Operation, CountryName, or Admin0Pcode.

        :param Operation: Name of the DTM operation for which the data was collected.
        :type Operation: Optional[str]
        :param CountryName: Name of the country where the data was collected.
        :type CountryName: Optional[str]
        :param Admin0Pcode: Country code (ISO 3166-1 alpha-3).
        :type Admin0Pcode: Optional[str]
        :param Admin1Name: Name of level 1 administrative boundaries.
        :type Admin1Name: Optional[str]
        :param Admin1Pcode: Place code of level 1 administrative boundaries.
        :type Admin1Pcode: Optional[str]
        :param FromReportingDate: Start date for the reporting period (format: 'YYYY-MM-DD').
        :type FromReportingDate: Optional[str]
        :param ToReportingDate: End date for the reporting period (format: 'YYYY-MM-DD').
        :type ToReportingDate: Optional[str]
        :param FromRoundNumber: Starting round number for the data collection range.
        :type FromRoundNumber: Optional[int]
        :param ToRoundNumber: Ending round number for the data collection range.
        :type ToRoundNumber: Optional[int]
        :param to_pandas: If True, the data will be returned as a pandas DataFrame. Otherwise, it will be returned as a JSON object.
        :type to_pandas: bool

        :return: The IDP Admin1 data matching the specified criteria, either as a DataFrame or a JSON object.
        :rtype: Union[pd.DataFrame, Dict[str, Any]]
        :raises ValidationError: If parameter validation fails.
        """
        # Validate required parameters
        validate_required_params(
            {"Operation": Operation, "CountryName": CountryName, "Admin0Pcode": Admin0Pcode},
            ["Operation", "CountryName", "Admin0Pcode"]
        )

        # Validate date formats
        if FromReportingDate:
            validate_date_format(FromReportingDate, "FromReportingDate")
        if ToReportingDate:
            validate_date_format(ToReportingDate, "ToReportingDate")
        validate_date_range(FromReportingDate, ToReportingDate, "FromReportingDate", "ToReportingDate")

        # Validate round numbers
        validate_round_number_range(FromRoundNumber, ToRoundNumber)

        params = {
            "Operation": Operation,
            "CountryName": CountryName,
            "Admin0Pcode": Admin0Pcode,
            "Admin1Name": Admin1Name,
            "Admin1Pcode": Admin1Pcode,
            "FromReportingDate": FromReportingDate,
            "ToReportingDate": ToReportingDate,
            "FromRoundNumber": FromRoundNumber,
            "ToRoundNumber": ToRoundNumber,
        }
        params = {k: v for k, v in params.items() if v is not None}
        return self._fetch_data(self._get_endpoint("admin1"), params, to_pandas)

    def get_idp_admin2_data(
        self,
        Operation: Optional[str] = None,
        CountryName: Optional[str] = None,
        Admin0Pcode: Optional[str] = None,
        Admin1Name: Optional[str] = None,
        Admin1Pcode: Optional[str] = None,
        Admin2Name: Optional[str] = None,
        Admin2Pcode: Optional[str] = None,
        FromReportingDate: Optional[str] = None,
        ToReportingDate: Optional[str] = None,
        FromRoundNumber: Optional[int] = None,
        ToRoundNumber: Optional[int] = None,
        to_pandas: bool = True,
    ) -> Union[pd.DataFrame, Dict[str, Any]]:
        """
        Retrieve IDP data at Admin 2 level based on specified parameters.

        At least one of the following parameters must be provided:
        Operation, CountryName, or Admin0Pcode.

        :param Operation: Name of the DTM operation for which the data was collected.
        :type Operation: Optional[str]
        :param CountryName: Name of the country where the data was collected.
        :type CountryName: Optional[str]
        :param Admin0Pcode: Country code (ISO 3166-1 alpha-3).
        :type Admin0Pcode: Optional[str]
        :param Admin1Name: Name of level 1 administrative boundaries.
        :type Admin1Name: Optional[str]
        :param Admin1Pcode: Place code of level 1 administrative boundaries.
        :type Admin1Pcode: Optional[str]
        :param Admin2Name: Name of level 2 administrative boundaries.
        :type Admin2Name: Optional[str]
        :param Admin2Pcode: Place code of level 2 administrative boundaries.
        :type Admin2Pcode: Optional[str]
        :param FromReportingDate: Start date for the reporting period (format: 'YYYY-MM-DD').
        :type FromReportingDate: Optional[str]
        :param ToReportingDate: End date for the reporting period (format: 'YYYY-MM-DD').
        :type ToReportingDate: Optional[str]
        :param FromRoundNumber: Starting round number for the data collection range.
        :type FromRoundNumber: Optional[int]
        :param ToRoundNumber: Ending round number for the data collection range.
        :type ToRoundNumber: Optional[int]
        :param to_pandas: If True, the data will be returned as a pandas DataFrame. Otherwise, it will be returned as a JSON object.
        :type to_pandas: bool

        :returns: The IDP Admin2 data matching the specified criteria, either as a DataFrame or a JSON object.
        :rtype: Union[pd.DataFrame, Dict[str, Any]]
        :raises ValidationError: If parameter validation fails.
        """
        # Validate required parameters
        validate_required_params(
            {"Operation": Operation, "CountryName": CountryName, "Admin0Pcode": Admin0Pcode},
            ["Operation", "CountryName", "Admin0Pcode"]
        )

        # Validate date formats
        if FromReportingDate:
            validate_date_format(FromReportingDate, "FromReportingDate")
        if ToReportingDate:
            validate_date_format(ToReportingDate, "ToReportingDate")
        validate_date_range(FromReportingDate, ToReportingDate, "FromReportingDate", "ToReportingDate")

        # Validate round numbers
        validate_round_number_range(FromRoundNumber, ToRoundNumber)

        params = {
            "Operation": Operation,
            "CountryName": CountryName,
            "Admin0Pcode": Admin0Pcode,
            "Admin1Name": Admin1Name,
            "Admin1Pcode": Admin1Pcode,
            "Admin2Name": Admin2Name,
            "Admin2Pcode": Admin2Pcode,
            "FromReportingDate": FromReportingDate,
            "ToReportingDate": ToReportingDate,
            "FromRoundNumber": FromRoundNumber,
            "ToRoundNumber": ToRoundNumber,
        }
        params = {k: v for k, v in params.items() if v is not None}
        return self._fetch_data(self._get_endpoint("admin2"), params, to_pandas)



"""Unit tests for DTMApi class."""

import os
import unittest
from unittest.mock import MagicMock, patch

import pandas as pd
import requests

from dtmapi import (
    DTMApi,
    DTMApiRequestError,
    DTMApiResponseError,
    DTMApiTimeoutError,
    DTMAuthenticationError,
)


class TestDTMApiInit(unittest.TestCase):
    """Test DTMApi initialization."""

    def test_init_with_subscription_key(self):
        """Test initialization with a subscription key."""
        api = DTMApi(subscription_key="test-key")
        self.assertEqual(api.subscription_key, "test-key")
        self.assertEqual(api.timeout, 30)

    def test_init_with_custom_timeout(self):
        """Test initialization with custom timeout."""
        api = DTMApi(subscription_key="test-key", timeout=60)
        self.assertEqual(api.timeout, 60)

    @patch.dict(os.environ, {"DTMAPI_SUBSCRIPTION_KEY": "env-key"})
    def test_init_with_env_var(self):
        """Test initialization with environment variable."""
        api = DTMApi()
        self.assertEqual(api.subscription_key, "env-key")

    def test_init_without_key_raises_error(self):
        """Test that initialization without key raises DTMAuthenticationError."""
        with patch.dict(os.environ, {}, clear=True):
            with self.assertRaises(DTMAuthenticationError) as context:
                DTMApi()
            self.assertIn("Subscription Key is required", str(context.exception))


class TestDTMApiHeaders(unittest.TestCase):
    """Test DTMApi headers generation."""

    def test_headers(self):
        """Test that headers are correctly generated."""
        api = DTMApi(subscription_key="test-key")
        headers = api._headers()
        self.assertEqual(headers["Ocp-Apim-Subscription-Key"], "test-key")
        self.assertIn("User-Agent", headers)


class TestDTMApiFetchData(unittest.TestCase):
    """Test DTMApi _fetch_data method."""

    def setUp(self):
        """Set up test fixtures."""
        self.api = DTMApi(subscription_key="test-key")

    @patch("requests.get")
    def test_fetch_data_success_with_pandas(self, mock_get):
        """Test successful data fetch returning pandas DataFrame."""
        mock_response = MagicMock()
        mock_response.json.return_value = {
            "isSuccess": True,
            "result": [{"country": "Ethiopia", "idps": 1000}],
        }
        mock_response.raise_for_status = MagicMock()
        mock_get.return_value = mock_response

        result = self.api._fetch_data("http://test.com/api", to_pandas=True)

        self.assertIsInstance(result, pd.DataFrame)
        self.assertEqual(len(result), 1)
        self.assertEqual(result.iloc[0]["country"], "Ethiopia")
        mock_get.assert_called_once()

    @patch("requests.get")
    def test_fetch_data_success_without_pandas(self, mock_get):
        """Test successful data fetch returning dict."""
        mock_response = MagicMock()
        mock_response.json.return_value = {
            "isSuccess": True,
            "result": [{"country": "Ethiopia", "idps": 1000}],
        }
        mock_response.raise_for_status = MagicMock()
        mock_get.return_value = mock_response

        result = self.api._fetch_data("http://test.com/api", to_pandas=False)

        self.assertIsInstance(result, list)
        self.assertEqual(len(result), 1)
        self.assertEqual(result[0]["country"], "Ethiopia")

    @patch("requests.get")
    def test_fetch_data_api_error(self, mock_get):
        """Test API error response."""
        mock_response = MagicMock()
        mock_response.json.return_value = {
            "isSuccess": False,
            "errorMessages": ["Invalid parameters"],
        }
        mock_response.raise_for_status = MagicMock()
        mock_get.return_value = mock_response

        with self.assertRaises(DTMApiResponseError) as context:
            self.api._fetch_data("http://test.com/api")
        self.assertEqual(str(context.exception), "Invalid parameters")
        self.assertEqual(context.exception.error_messages, ["Invalid parameters"])

    @patch("requests.get")
    def test_fetch_data_api_error_empty_messages(self, mock_get):
        """Test API error with empty error messages."""
        mock_response = MagicMock()
        mock_response.json.return_value = {"isSuccess": False, "errorMessages": []}
        mock_response.raise_for_status = MagicMock()
        mock_get.return_value = mock_response

        with self.assertRaises(DTMApiResponseError) as context:
            self.api._fetch_data("http://test.com/api")
        self.assertEqual(str(context.exception), "Unknown API error")

    @patch("requests.get")
    def test_fetch_data_timeout(self, mock_get):
        """Test timeout error."""
        mock_get.side_effect = requests.Timeout("Connection timeout")

        with self.assertRaises(DTMApiTimeoutError) as context:
            self.api._fetch_data("http://test.com/api")
        self.assertIn("timed out after 30 seconds", str(context.exception))

    @patch("requests.get")
    def test_fetch_data_authentication_error_401(self, mock_get):
        """Test authentication error with 401 status."""
        mock_response = MagicMock()
        mock_response.status_code = 401
        mock_response.reason = "Unauthorized"
        http_error = requests.HTTPError()
        http_error.response = mock_response
        mock_get.side_effect = http_error

        with self.assertRaises(DTMAuthenticationError) as context:
            self.api._fetch_data("http://test.com/api")
        self.assertIn("401", str(context.exception))

    @patch("requests.get")
    def test_fetch_data_authentication_error_403(self, mock_get):
        """Test authentication error with 403 status."""
        mock_response = MagicMock()
        mock_response.status_code = 403
        mock_response.reason = "Forbidden"
        http_error = requests.HTTPError()
        http_error.response = mock_response
        mock_get.side_effect = http_error

        with self.assertRaises(DTMAuthenticationError) as context:
            self.api._fetch_data("http://test.com/api")
        self.assertIn("403", str(context.exception))

    @patch("requests.get")
    def test_fetch_data_http_error(self, mock_get):
        """Test HTTP error (non-auth)."""
        mock_response = MagicMock()
        mock_response.status_code = 500
        mock_response.reason = "Internal Server Error"
        http_error = requests.HTTPError()
        http_error.response = mock_response
        mock_get.side_effect = http_error

        with self.assertRaises(DTMApiRequestError) as context:
            self.api._fetch_data("http://test.com/api")
        self.assertIn("500", str(context.exception))

    @patch("requests.get")
    def test_fetch_data_request_error(self, mock_get):
        """Test general request error."""
        mock_get.side_effect = requests.ConnectionError("Network error")

        with self.assertRaises(DTMApiRequestError) as context:
            self.api._fetch_data("http://test.com/api")
        self.assertIn("API request failed", str(context.exception))

    @patch("requests.get")
    def test_fetch_data_with_timeout_parameter(self, mock_get):
        """Test that timeout parameter is passed to requests.get."""
        mock_response = MagicMock()
        mock_response.json.return_value = {"isSuccess": True, "result": []}
        mock_response.raise_for_status = MagicMock()
        mock_get.return_value = mock_response

        api = DTMApi(subscription_key="test-key", timeout=60)
        api._fetch_data("http://test.com/api")

        # Verify timeout was passed
        call_kwargs = mock_get.call_args[1]
        self.assertEqual(call_kwargs["timeout"], 60)


class TestDTMApiPublicMethods(unittest.TestCase):
    """Test DTMApi public methods."""

    def setUp(self):
        """Set up test fixtures."""
        self.api = DTMApi(subscription_key="test-key")

    @patch("dtmapi.api.DTMApi._fetch_data")
    def test_get_all_countries(self, mock_fetch):
        """Test get_all_countries method."""
        mock_fetch.return_value = pd.DataFrame([{"country": "Ethiopia"}])
        result = self.api.get_all_countries()
        self.assertIsInstance(result, pd.DataFrame)
        mock_fetch.assert_called_once()

    @patch("dtmapi.api.DTMApi._fetch_data")
    def test_get_all_operations(self, mock_fetch):
        """Test get_all_operations method."""
        mock_fetch.return_value = pd.DataFrame([{"operation": "IDP Tracking"}])
        result = self.api.get_all_operations()
        self.assertIsInstance(result, pd.DataFrame)
        mock_fetch.assert_called_once()

    @patch("dtmapi.api.DTMApi._fetch_data")
    def test_get_idp_admin0_data(self, mock_fetch):
        """Test get_idp_admin0_data method."""
        mock_fetch.return_value = pd.DataFrame([{"country": "Ethiopia", "idps": 1000}])
        result = self.api.get_idp_admin0_data(CountryName="Ethiopia")
        self.assertIsInstance(result, pd.DataFrame)

        # Verify params were filtered correctly
        call_args = mock_fetch.call_args
        params = call_args[0][1]
        self.assertEqual(params["CountryName"], "Ethiopia")
        self.assertNotIn("Operation", params)

    @patch("dtmapi.api.DTMApi._fetch_data")
    def test_get_idp_admin1_data(self, mock_fetch):
        """Test get_idp_admin1_data method."""
        mock_fetch.return_value = pd.DataFrame([{"admin1": "Tigray", "idps": 500}])
        result = self.api.get_idp_admin1_data(
            CountryName="Ethiopia", Admin1Name="Tigray"
        )
        self.assertIsInstance(result, pd.DataFrame)

        # Verify params were filtered correctly
        call_args = mock_fetch.call_args
        params = call_args[0][1]
        self.assertEqual(params["CountryName"], "Ethiopia")
        self.assertEqual(params["Admin1Name"], "Tigray")

    @patch("dtmapi.api.DTMApi._fetch_data")
    def test_get_idp_admin2_data(self, mock_fetch):
        """Test get_idp_admin2_data method."""
        mock_fetch.return_value = pd.DataFrame([{"admin2": "District1", "idps": 200}])
        result = self.api.get_idp_admin2_data(
            CountryName="Ethiopia", Admin2Name="District1"
        )
        self.assertIsInstance(result, pd.DataFrame)

        # Verify params were filtered correctly
        call_args = mock_fetch.call_args
        params = call_args[0][1]
        self.assertEqual(params["CountryName"], "Ethiopia")
        self.assertEqual(params["Admin2Name"], "District1")

    @patch("dtmapi.api.DTMApi._fetch_data")
    def test_get_idp_admin0_data_filters_none_values(self, mock_fetch):
        """Test that None values are filtered from params."""
        mock_fetch.return_value = pd.DataFrame([])
        self.api.get_idp_admin0_data(
            CountryName="Ethiopia",
            Operation=None,
            FromRoundNumber=1,
            ToRoundNumber=None,
        )

        call_args = mock_fetch.call_args
        params = call_args[0][1]
        self.assertIn("CountryName", params)
        self.assertIn("FromRoundNumber", params)
        self.assertNotIn("Operation", params)
        self.assertNotIn("ToRoundNumber", params)


class TestDTMApiRetryLogic(unittest.TestCase):
    """Test DTMApi retry logic."""

    def setUp(self):
        """Set up test fixtures."""
        self.api = DTMApi(subscription_key="test-key", max_retries=2, retry_delay=0.1)

    def test_init_with_retry_params(self):
        """Test initialization with retry parameters."""
        api = DTMApi(subscription_key="test-key", max_retries=5, retry_delay=2.0)
        self.assertEqual(api.max_retries, 5)
        self.assertEqual(api.retry_delay, 2.0)

    def test_init_default_retry_params(self):
        """Test default retry parameters."""
        api = DTMApi(subscription_key="test-key")
        self.assertEqual(api.max_retries, 3)
        self.assertEqual(api.retry_delay, 1)

    def test_is_retryable_error_timeout(self):
        """Test that timeout errors are retryable."""
        error = requests.Timeout("Connection timeout")
        self.assertTrue(self.api._is_retryable_error(error))

    def test_is_retryable_error_connection(self):
        """Test that connection errors are retryable."""
        error = requests.ConnectionError("Connection failed")
        self.assertTrue(self.api._is_retryable_error(error))

    def test_is_retryable_error_http_429(self):
        """Test that 429 errors are retryable."""
        mock_response = MagicMock()
        mock_response.status_code = 429
        error = requests.HTTPError()
        error.response = mock_response
        self.assertTrue(self.api._is_retryable_error(error))

    def test_is_retryable_error_http_500(self):
        """Test that 500 errors are retryable."""
        mock_response = MagicMock()
        mock_response.status_code = 500
        error = requests.HTTPError()
        error.response = mock_response
        self.assertTrue(self.api._is_retryable_error(error))

    def test_is_not_retryable_error_http_404(self):
        """Test that 404 errors are not retryable."""
        mock_response = MagicMock()
        mock_response.status_code = 404
        error = requests.HTTPError()
        error.response = mock_response
        self.assertFalse(self.api._is_retryable_error(error))

    @patch("dtmapi.api.time.sleep")
    @patch("requests.get")
    def test_retry_on_timeout(self, mock_get, mock_sleep):
        """Test that request is retried on timeout."""
        mock_response = MagicMock()
        mock_response.json.return_value = {"isSuccess": True, "result": []}
        mock_response.raise_for_status = MagicMock()

        mock_get.side_effect = [
            requests.Timeout("Timeout 1"),
            requests.Timeout("Timeout 2"),
            mock_response,
        ]

        result = self.api._fetch_data("http://test.com/api")

        self.assertEqual(mock_get.call_count, 3)
        self.assertEqual(mock_sleep.call_count, 2)

    @patch("dtmapi.api.time.sleep")
    @patch("requests.get")
    def test_retry_exhaustion(self, mock_get, mock_sleep):
        """Test that retries are exhausted and error is raised."""
        mock_get.side_effect = requests.Timeout("Persistent timeout")

        with self.assertRaises(DTMApiTimeoutError):
            self.api._fetch_data("http://test.com/api")

        self.assertEqual(mock_get.call_count, 3)

    @patch("requests.get")
    def test_no_retry_on_auth_error(self, mock_get):
        """Test that authentication errors are not retried."""
        mock_response = MagicMock()
        mock_response.status_code = 401
        mock_response.reason = "Unauthorized"
        error = requests.HTTPError()
        error.response = mock_response
        mock_get.side_effect = error

        with self.assertRaises(DTMAuthenticationError):
            self.api._fetch_data("http://test.com/api")

        self.assertEqual(mock_get.call_count, 1)


class TestDTMApiValidation(unittest.TestCase):
    """Test DTMApi parameter validation."""

    def setUp(self):
        """Set up test fixtures."""
        self.api = DTMApi(subscription_key="test-key")

    def test_missing_required_params(self):
        """Test that missing required params raises ValidationError."""
        from dtmapi import ValidationError
        with self.assertRaises(ValidationError) as context:
            self.api.get_idp_admin0_data()
        self.assertIn("At least one", str(context.exception))

    def test_invalid_date_format(self):
        """Test that invalid date format raises ValidationError."""
        from dtmapi import ValidationError
        with self.assertRaises(ValidationError) as context:
            self.api.get_idp_admin0_data(
                CountryName="Ethiopia",
                FromReportingDate="01-01-2024"
            )
        self.assertIn("YYYY-MM-DD", str(context.exception))

    def test_invalid_date_range(self):
        """Test that invalid date range raises ValidationError."""
        from dtmapi import ValidationError
        with self.assertRaises(ValidationError) as context:
            self.api.get_idp_admin0_data(
                CountryName="Ethiopia",
                FromReportingDate="2024-12-31",
                ToReportingDate="2024-01-01"
            )
        self.assertIn("must be before or equal to", str(context.exception))

    def test_invalid_round_number(self):
        """Test that invalid round number raises ValidationError."""
        from dtmapi import ValidationError
        with self.assertRaises(ValidationError) as context:
            self.api.get_idp_admin0_data(
                CountryName="Ethiopia",
                FromRoundNumber=-1
            )
        self.assertIn("must be >= 1", str(context.exception))

    def test_invalid_round_range(self):
        """Test that invalid round range raises ValidationError."""
        from dtmapi import ValidationError
        with self.assertRaises(ValidationError) as context:
            self.api.get_idp_admin0_data(
                CountryName="Ethiopia",
                FromRoundNumber=10,
                ToRoundNumber=1
            )
        self.assertIn("must be <=", str(context.exception))

    @patch("dtmapi.api.DTMApi._fetch_data")
    def test_valid_params_pass_validation(self, mock_fetch):
        """Test that valid params pass validation."""
        mock_fetch.return_value = pd.DataFrame([])

        self.api.get_idp_admin0_data(
            CountryName="Ethiopia",
            FromReportingDate="2024-01-01",
            ToReportingDate="2024-12-31",
            FromRoundNumber=1,
            ToRoundNumber=10
        )

        mock_fetch.assert_called_once()


class TestDTMApiVersions(unittest.TestCase):
    """Test DTMApi version selection and endpoint routing."""

    def test_init_with_v3_version(self):
        """Test initialization with v3 version."""
        api = DTMApi(subscription_key="test-key", api_version="v3")
        self.assertEqual(api.api_version, "v3")

    def test_init_with_v2_version(self):
        """Test initialization with v2 version."""
        api = DTMApi(subscription_key="test-key", api_version="v2")
        self.assertEqual(api.api_version, "v2")

    def test_init_default_version_is_v3(self):
        """Test that default version is v3."""
        api = DTMApi(subscription_key="test-key")
        self.assertEqual(api.api_version, "v3")

    def test_init_invalid_version_raises_error(self):
        """Test that invalid version raises DTMApiVersionError."""
        from dtmapi import DTMApiVersionError
        with self.assertRaises(DTMApiVersionError) as context:
            DTMApi(subscription_key="test-key", api_version="v1")
        self.assertIn("must be 'v2' or 'v3'", str(context.exception))

    def test_get_endpoint_v3_admin0(self):
        """Test endpoint selection for v3 admin0."""
        api = DTMApi(subscription_key="test-key", api_version="v3")
        endpoint = api._get_endpoint("admin0")
        self.assertIn("/v3/displacement/admin0", endpoint)

    def test_get_endpoint_v2_admin0(self):
        """Test endpoint selection for v2 admin0."""
        api = DTMApi(subscription_key="test-key", api_version="v2")
        endpoint = api._get_endpoint("admin0")
        self.assertIn("/v2/IdpAdmin0Data", endpoint)

    def test_get_endpoint_v3_admin1(self):
        """Test endpoint selection for v3 admin1."""
        api = DTMApi(subscription_key="test-key", api_version="v3")
        endpoint = api._get_endpoint("admin1")
        self.assertIn("/v3/displacement/admin1", endpoint)

    def test_get_endpoint_v2_admin1(self):
        """Test endpoint selection for v2 admin1."""
        api = DTMApi(subscription_key="test-key", api_version="v2")
        endpoint = api._get_endpoint("admin1")
        self.assertIn("/v2/IdpAdmin1Data", endpoint)

    def test_get_endpoint_v3_admin2(self):
        """Test endpoint selection for v3 admin2."""
        api = DTMApi(subscription_key="test-key", api_version="v3")
        endpoint = api._get_endpoint("admin2")
        self.assertIn("/v3/displacement/admin2", endpoint)

    def test_get_endpoint_v2_admin2(self):
        """Test endpoint selection for v2 admin2."""
        api = DTMApi(subscription_key="test-key", api_version="v2")
        endpoint = api._get_endpoint("admin2")
        self.assertIn("/v2/IdpAdmin2Data", endpoint)

    def test_get_endpoint_v3_countries(self):
        """Test endpoint selection for v3 countries."""
        api = DTMApi(subscription_key="test-key", api_version="v3")
        endpoint = api._get_endpoint("countries")
        self.assertIn("/v3/displacement/country-list", endpoint)

    def test_get_endpoint_v2_countries(self):
        """Test endpoint selection for v2 countries."""
        api = DTMApi(subscription_key="test-key", api_version="v2")
        endpoint = api._get_endpoint("countries")
        self.assertIn("/v2/CountryList", endpoint)

    def test_get_endpoint_v3_operations(self):
        """Test endpoint selection for v3 operations."""
        api = DTMApi(subscription_key="test-key", api_version="v3")
        endpoint = api._get_endpoint("operations")
        self.assertIn("/v3/displacement/operation-list", endpoint)

    def test_get_endpoint_v2_operations(self):
        """Test endpoint selection for v2 operations."""
        api = DTMApi(subscription_key="test-key", api_version="v2")
        endpoint = api._get_endpoint("operations")
        self.assertIn("/v2/OperationList", endpoint)

    @patch("dtmapi.api.DTMApi._fetch_data")
    def test_v2_idp_admin0_uses_correct_endpoint(self, mock_fetch):
        """Test that v2 get_idp_admin0_data uses v2 endpoint."""
        mock_fetch.return_value = pd.DataFrame([])
        api = DTMApi(subscription_key="test-key", api_version="v2")
        api.get_idp_admin0_data(CountryName="Ethiopia")

        # Verify the endpoint URL contains v2
        call_args = mock_fetch.call_args
        endpoint_url = call_args[0][0]
        self.assertIn("/v2/IdpAdmin0Data", endpoint_url)

    @patch("dtmapi.api.DTMApi._fetch_data")
    def test_v3_idp_admin0_uses_correct_endpoint(self, mock_fetch):
        """Test that v3 get_idp_admin0_data uses v3 endpoint."""
        mock_fetch.return_value = pd.DataFrame([])
        api = DTMApi(subscription_key="test-key", api_version="v3")
        api.get_idp_admin0_data(CountryName="Ethiopia")

        # Verify the endpoint URL contains v3
        call_args = mock_fetch.call_args
        endpoint_url = call_args[0][0]
        self.assertIn("/v3/displacement/admin0", endpoint_url)

    @patch("dtmapi.api.DTMApi._fetch_data")
    def test_v2_countries_uses_correct_endpoint(self, mock_fetch):
        """Test that v2 get_all_countries uses v2 endpoint."""
        mock_fetch.return_value = pd.DataFrame([])
        api = DTMApi(subscription_key="test-key", api_version="v2")
        api.get_all_countries()

        # Verify the endpoint URL contains v2
        call_args = mock_fetch.call_args
        endpoint_url = call_args[0][0]
        self.assertIn("/v2/CountryList", endpoint_url)

    @patch("dtmapi.api.DTMApi._fetch_data")
    def test_v3_operations_uses_correct_endpoint(self, mock_fetch):
        """Test that v3 get_all_operations uses v3 endpoint."""
        mock_fetch.return_value = pd.DataFrame([])
        api = DTMApi(subscription_key="test-key", api_version="v3")
        api.get_all_operations()

        # Verify the endpoint URL contains v3
        call_args = mock_fetch.call_args
        endpoint_url = call_args[0][0]
        self.assertIn("/v3/displacement/operation-list", endpoint_url)


if __name__ == "__main__":
    unittest.main()



"""Input validation utilities for DTM API."""

import re
from datetime import datetime
from typing import Any, Dict, List, Optional


class ValidationError(ValueError):
    """Raised when input validation fails."""

    pass


def validate_date_format(date_string: str, param_name: str) -> None:
    """
    Validate that a date string is in YYYY-MM-DD format.

    :param date_string: The date string to validate.
    :type date_string: str
    :param param_name: The parameter name for error messages.
    :type param_name: str
    :raises ValidationError: If date format is invalid.
    """
    if not date_string:
        return

    date_pattern = re.compile(r"^\d{4}-\d{2}-\d{2}$")
    if not date_pattern.match(date_string):
        raise ValidationError(
            f"{param_name} must be in YYYY-MM-DD format, got: {date_string}"
        )

    try:
        datetime.strptime(date_string, "%Y-%m-%d")
    except ValueError as e:
        raise ValidationError(f"{param_name} is not a valid date: {e}") from e


def validate_date_range(
    from_date: Optional[str], to_date: Optional[str], from_param: str, to_param: str
) -> None:
    """
    Validate that date range is logical (from_date <= to_date).

    :param from_date: Start date string.
    :type from_date: Optional[str]
    :param to_date: End date string.
    :type to_date: Optional[str]
    :param from_param: Name of the from parameter.
    :type from_param: str
    :param to_param: Name of the to parameter.
    :type to_param: str
    :raises ValidationError: If date range is invalid.
    """
    if not from_date or not to_date:
        return

    from_dt = datetime.strptime(from_date, "%Y-%m-%d")
    to_dt = datetime.strptime(to_date, "%Y-%m-%d")

    if from_dt > to_dt:
        raise ValidationError(
            f"{from_param} ({from_date}) must be before or equal to {to_param} ({to_date})"
        )


def validate_round_number_range(
    from_round: Optional[int], to_round: Optional[int]
) -> None:
    """
    Validate round number range.

    :param from_round: Starting round number.
    :type from_round: Optional[int]
    :param to_round: Ending round number.
    :type to_round: Optional[int]
    :raises ValidationError: If range is invalid.
    """
    if from_round is not None and from_round < 1:
        raise ValidationError(f"FromRoundNumber must be >= 1, got: {from_round}")

    if to_round is not None and to_round < 1:
        raise ValidationError(f"ToRoundNumber must be >= 1, got: {to_round}")

    if from_round is not None and to_round is not None and from_round > to_round:
        raise ValidationError(
            f"FromRoundNumber ({from_round}) must be <= ToRoundNumber ({to_round})"
        )


def validate_required_params(
    params: Dict[str, Any], required_one_of: List[str]
) -> None:
    """
    Validate that at least one of the required parameters is provided.

    :param params: Dictionary of parameters.
    :type params: Dict[str, Any]
    :param required_one_of: List of parameter names where at least one is required.
    :type required_one_of: List[str]
    :raises ValidationError: If no required parameters are provided.
    """
    if not any(params.get(param) is not None for param in required_one_of):
        raise ValidationError(
            f"At least one of the following parameters is required: {', '.join(required_one_of)}"
        )



"""Unit tests for validators."""

import unittest

from dtmapi.validators import (
    ValidationError,
    validate_date_format,
    validate_date_range,
    validate_required_params,
    validate_round_number_range,
)


class TestValidateDateFormat(unittest.TestCase):
    """Test date format validation."""

    def test_valid_date(self):
        """Test validation with valid date."""
        validate_date_format("2024-01-15", "TestDate")  # Should not raise

    def test_invalid_format(self):
        """Test validation with invalid format."""
        with self.assertRaises(ValidationError) as context:
            validate_date_format("01-15-2024", "TestDate")
        self.assertIn("YYYY-MM-DD", str(context.exception))

    def test_invalid_date(self):
        """Test validation with invalid date."""
        with self.assertRaises(ValidationError) as context:
            validate_date_format("2024-13-01", "TestDate")
        self.assertIn("not a valid date", str(context.exception))

    def test_empty_string(self):
        """Test validation with empty string."""
        validate_date_format("", "TestDate")  # Should not raise


class TestValidateDateRange(unittest.TestCase):
    """Test date range validation."""

    def test_valid_range(self):
        """Test validation with valid range."""
        validate_date_range("2024-01-01", "2024-12-31", "FromDate", "ToDate")  # Should not raise

    def test_same_dates(self):
        """Test validation with same dates."""
        validate_date_range("2024-01-01", "2024-01-01", "FromDate", "ToDate")  # Should not raise

    def test_invalid_range(self):
        """Test validation with invalid range (from > to)."""
        with self.assertRaises(ValidationError) as context:
            validate_date_range("2024-12-31", "2024-01-01", "FromDate", "ToDate")
        self.assertIn("must be before or equal to", str(context.exception))

    def test_none_from_date(self):
        """Test validation with None from date."""
        validate_date_range(None, "2024-12-31", "FromDate", "ToDate")  # Should not raise

    def test_none_to_date(self):
        """Test validation with None to date."""
        validate_date_range("2024-01-01", None, "FromDate", "ToDate")  # Should not raise

    def test_both_none(self):
        """Test validation with both dates None."""
        validate_date_range(None, None, "FromDate", "ToDate")  # Should not raise


class TestValidateRoundNumberRange(unittest.TestCase):
    """Test round number range validation."""

    def test_valid_range(self):
        """Test validation with valid range."""
        validate_round_number_range(1, 10)  # Should not raise

    def test_same_numbers(self):
        """Test validation with same numbers."""
        validate_round_number_range(5, 5)  # Should not raise

    def test_invalid_range(self):
        """Test validation with invalid range (from > to)."""
        with self.assertRaises(ValidationError) as context:
            validate_round_number_range(10, 1)
        self.assertIn("must be <=", str(context.exception))

    def test_negative_from_round(self):
        """Test validation with negative from round."""
        with self.assertRaises(ValidationError) as context:
            validate_round_number_range(-1, 10)
        self.assertIn("must be >= 1", str(context.exception))

    def test_zero_from_round(self):
        """Test validation with zero from round."""
        with self.assertRaises(ValidationError) as context:
            validate_round_number_range(0, 10)
        self.assertIn("must be >= 1", str(context.exception))

    def test_negative_to_round(self):
        """Test validation with negative to round."""
        with self.assertRaises(ValidationError) as context:
            validate_round_number_range(1, -1)
        self.assertIn("must be >= 1", str(context.exception))

    def test_none_from_round(self):
        """Test validation with None from round."""
        validate_round_number_range(None, 10)  # Should not raise

    def test_none_to_round(self):
        """Test validation with None to round."""
        validate_round_number_range(1, None)  # Should not raise

    def test_both_none(self):
        """Test validation with both None."""
        validate_round_number_range(None, None)  # Should not raise


class TestValidateRequiredParams(unittest.TestCase):
    """Test required parameters validation."""

    def test_one_param_present(self):
        """Test validation with one required param present."""
        params = {"Operation": "IDP", "CountryName": None, "Admin0Pcode": None}
        validate_required_params(params, ["Operation", "CountryName", "Admin0Pcode"])  # Should not raise

    def test_all_params_present(self):
        """Test validation with all params present."""
        params = {"Operation": "IDP", "CountryName": "Ethiopia", "Admin0Pcode": "ETH"}
        validate_required_params(params, ["Operation", "CountryName", "Admin0Pcode"])  # Should not raise

    def test_no_params_present(self):
        """Test validation with no required params present."""
        params = {"Operation": None, "CountryName": None, "Admin0Pcode": None}
        with self.assertRaises(ValidationError) as context:
            validate_required_params(params, ["Operation", "CountryName", "Admin0Pcode"])
        self.assertIn("At least one", str(context.exception))

    def test_empty_dict(self):
        """Test validation with empty dict."""
        params = {}
        with self.assertRaises(ValidationError) as context:
            validate_required_params(params, ["Operation", "CountryName"])
        self.assertIn("At least one", str(context.exception))


if __name__ == "__main__":
    unittest.main()



from .api import DTMApi
from .exceptions import (
    DTMApiError,
    DTMApiRequestError,
    DTMApiResponseError,
    DTMApiTimeoutError,
    DTMAuthenticationError,
    DTMApiVersionError,
)
from .validators import ValidationError
from .version import __version__

__all__ = [
    "DTMApi",
    "DTMApiError",
    "DTMApiRequestError",
    "DTMApiResponseError",
    "DTMApiTimeoutError",
    "DTMAuthenticationError",
    "DTMApiVersionError",
    "ValidationError",
    "__version__",
]



# setup.py

from setuptools import find_packages, setup

# Read version from version.py
with open("dtmapi/version.py", "r") as f:
    exec(f.read())

setup(
    name="dtmapi",
    version=__version__,  # noqa: F821
    packages=find_packages(),
    install_requires=[
        "requests>=2.31.0",
        "pandas>=2.0.0",
        "openpyxl>=3.1.0",
    ],
    extras_require={
        "test": [
            "pytest>=9.0.0",
            "pytest-cov>=7.0.0",
            "pytest-mock>=3.15.1",
        ],
        "docs": [
            "sphinx>=7.0.0",
            "sphinx-rtd-theme>=2.0.0",
        ],
        "dev": [
            "ipykernel>=7.1.0",
        ],
    },
    description="A Python package for fetching data from the IOM's Displacement Tracking Matrix (DTM).",
    author="Luong Bang Tran",
    author_email="lutran@iom.int",
    license="MIT",
    url="https://github.com/Displacement-Tracking-Matrix/dtmapi",
    keywords=["dtm", "displacement", "tracking", "matrix", "iom", "idp", "humanitarian", "migration", "refugees"],
    long_description=open("README.md", encoding="utf-8").read(),
    long_description_content_type="text/markdown",
    project_urls={
        "Github": "https://github.com/Displacement-Tracking-Matrix/dtmapi",
        "Documentation": "https://dtmapi.readthedocs.io/en/latest/",
    },
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering",
        "Topic :: Sociology",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.8",
)
